#include "UnstructuredGrid3D.h"
#include "PointSearch.h"
#include "MeshPCA.h"
#include "pca.h"

using namespace std;



inline float G(float x)												//Momentum classification function (Clarenz, 2004)
{
	const float alpha = 0.01;
	const float beta  = 100;
	return beta - 1/(alpha + beta*x*x);
}	 
	

static void plane(const std::vector<Point3d>& pts,float& variation, float& planarity, float& classifier, float radius)
{																	//Given a set of 3D points, compute the eigenvectors and
																	//eigenvalues of the points' covariance matrix.
																	//Next, compute three surface classifiers based on this tensor
    float *M[3],*V[3],W[3],MM0[3],MM1[3],MM2[3],VV0[3],VV1[3],VV2[3],m[6]; 
    int N = pts.size();
  
    M[0] = MM0; M[1] = MM1; M[2] = MM2; V[0] = VV0; V[1] = VV1; V[2] = VV2; 

    Point3d p;														//Compute p = barycenter of neighborhood
    for(int i=0;i<N;++i)											
	   p += pts[i];	
    p /= N;															

    m[0] = m[1] = m[2] = m[3] = m[4] = m[5] = 0;					//Compute m[] = 1st moment
    for(int i=0;i<N;++i)												
    {
	   Point3d d = pts[i]-p;	
       m[0] += d[0]*d[0];  m[1] += d[0]*d[1]; m[2] += d[0]*d[2];	
       m[3] += d[1]*d[1];  m[4] += d[1]*d[2]; m[5] += d[2]*d[2];
    }
	
	for(int i=0;i<6;++i) m[i] /= N;

    M[0][0] = m[0]; M[0][1] = m[1]; M[0][2] = m[2];					//Fill-in symmetric M[][] for eigen-computation
    M[1][0] = m[1]; M[1][1] = m[3]; M[1][2] = m[4];
    M[2][0] = m[2]; M[2][1] = m[4]; M[2][2] = m[5];
    
    JacobiN(M,3,W,V);												//Compute the eigenvalues W[] and vectors V[] of M[][]
																	//Results are returned in decreasing eigenval-order W[0]>W[1]>W[2]
    W[0] = fabs(W[0]); W[1] = fabs(W[1]); W[2] = fabs(W[2]);
		
	float shift = (pts[0]-p).norm();								//Distance between current point and neighborhood barycenter
		
	variation  = W[2]/(W[0]+W[1]+W[2]);								//Pauly's surface variation metric [Pauly, 2003]
	planarity  = 1 - 2*(W[1]-W[2])/(W[0]+W[1]+W[2]);				//Westin's planar anisotropy metric [Westin, 1997]
	classifier = G(shift*W[2]/W[0]/radius);							//Clarenz's moment classifier metric [Clarenz, 2004]
}




MeshPCA::MeshPCA(UnstructuredGrid3D& grid_,int numNeighbors_)
	   : grid(grid_),
	     variation(grid_.numPoints()), planarity(grid_.numPoints()), classifier(grid_.numPoints()),
		 numNeighbors(numNeighbors_),
	     search_on_surface(false)	 
{
																	//  Initialize a spatial search engine on the cloud points. We'll use
																	//  this next to find point-neighborhoods to estimate the	
   point_search = new PointSearch(grid,numNeighbors);				//  surface's curvature tensor
}




MeshPCA::~MeshPCA()
{
	delete point_search;
}


void MeshPCA::compute(float radius)									//Compute the curvature tensor for neighborhoods of size 'radius'
{
   vector<Point3d> nbs; nbs.reserve(numNeighbors);	  

   variation.resetRange(); planarity.resetRange(); classifier.resetRange();				
																	//Since we next fill in all these arrays, reset their min/max ranges
   
   VectorAttributes& normals = grid.getPointNormals();
   
   vector<int>   indexes;
   vector<float> distances;											//
   indexes.reserve(numNeighbors);		
	
   for(int pid=0;pid<grid.numPoints();++pid)						//1. Estimate the curvature tensor and classification metrics
   {																//   at each surface point 'pid'
	  Point3d p;
	  grid.getPoint(pid,p.data);	  
	  nbs.clear();								
	  
	  if (search_on_surface)										//1.1. Search neighborhood of 'pid' either on mesh surface
		 grid.closest(radius,pid,numNeighbors,indexes);				//     (slower, but more accurate), or in 3D space 
	  else															//     (faster, but less accurate for thin object areas)
	     point_search->closest(radius,p,indexes,distances);			
	  
	  int NN = indexes.size();										//NN = actual number of neighbors returned (<=numNeighbors)
	  for(int i=0;i<NN;++i)											//1.2. Collect all NN neighbors in nbs[]
	  {
	    Point3d pi;
		grid.getPoint(indexes[i],pi);
		nbs.push_back(pi);
	  }
	  												
	  if (nbs.size()>=3)											//1.2. Compute PCA only if there are enough neighbors
	  {
		float L0,L1,L2;
	    plane(nbs,L0,L1,L2,radius);									//1.3. Perform the PCA on the neighborhood nbs[]
		
		variation.setC0Scalar(pid,L0);								//1.4. Store the classifiers computed for point 'pid'
		planarity.setC0Scalar(pid,L1);
		classifier.setC0Scalar(pid,L2);
	  }
   }
}


