#include "MeshRenderer.h"
#include "UnstructuredGrid3D.h"
#include <GLUT/glut.h>									//GLUT library
#include <math.h>
#include <algorithm>

using namespace std;



void rainbow(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using the rainbow colormap
{
   const float dx=0.8f;

   value = (6-2*dx)*value+dx;
   R = max(0.0f,(3-(float)fabs(value-4)-(float)fabs(value-5))/2);
   G = max(0.0f,(4-(float)fabs(value-2)-(float)fabs(value-4))/2);
   B = max(0.0f,(3-(float)fabs(value-1)-(float)fabs(value-2))/2);
}



void MeshRenderer::createTextures()						//Create one 1D texture to encode our rainbow colormap.
{														//We will next use this texture to color map scalar data.

	glGenTextures(1,&textureID);						//Generate a texture names, for the texture we will create
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);				//Make sure that OpenGL will understand our CPU-side texture storage format

	glBindTexture(GL_TEXTURE_1D,textureID);				//Make our texture active (for setting it)
	const int size = 512;								//Allocate a texture-buffer large enough to store our colormap with high resolution
	float textureImage[3*size];

	for(int j=0;j<size;++j)								//Generate all 'size' RGB texels for the current texture:
	{
		float v = float(j)/(size-1);					//Compute a scalar value in [0,1]
		float R,G,B;
		rainbow(v,R,G,B);								//Map this scalar value to a color, using our colormap
	
		textureImage[3*j]   = R;						//Store the color for this scalar value in the texture
		textureImage[3*j+1] = G;
		textureImage[3*j+2] = B;
	}	
	glTexImage1D(GL_TEXTURE_1D,0,GL_RGB,size,0,GL_RGB,GL_FLOAT,textureImage);
														//The texture is ready - pass it to OpenGL
}




void MeshRenderer::draw(Grid& g_)						//Draw an unstructured grid, in the style indicated by 'drawing_style'
{
	glClearColor(1,1,1,1);								//1. Clear the frame and depth buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);								//2. Ask OpenGL to automatically normalize normal-vectors when we scale the model
														//   This is required to get correct shading for various model scales.

	UnstructuredGrid3D& g = (UnstructuredGrid3D&)g_;
	
	switch (draw_style)									//3. Render the grid, based on the current drawing style
	{
	case DRAW_GRID:
		glColor3f(1,1,1);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);				//   Use OpenGL polygon offsetting to draw the filled polygon mesh a bit 	
		glPolygonOffset(1.0,1.0);						//   behind the wireframe drawing. This makes the two overlay nicely.
		drawGrid(g);
		glDisable(GL_POLYGON_OFFSET_FILL);
		glColor3f(0,0,0);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		drawGrid(g);
		break;
	case DRAW_C1_CELLS:									 
		drawC1Cells(g);
		if (show_vectors) drawVectors(g);				//   Show eigenvectors (if glyphs enabled)
		break;
	case DRAW_C1_TEX_CELLS:
		drawC1TexCells(g);
		if (show_vectors) drawVectors(g);				//   Show eigenvectors (if glyphs enabled)
		break;
	}			
	
	glutSwapBuffers();
}




void MeshRenderer::drawGrid(UnstructuredGrid3D& g)		//Draw 'g' without scalars or vectors
{	
	glEnable(GL_COLOR_MATERIAL);
  	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glShadeModel(GL_SMOOTH);

	VectorAttributes& point_normals = g.getPointNormals();

	for(int i=0;i<g.numCells();++i)						//1. Draw all cells in the grid
	{
		int   cell[10];									//
		
		int nvertices = g.getCell(i,cell);				//2.   Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:		

		for(int j=0;j<nvertices;++j)					//3.   Render current cell
		{
			float x[3];
			g.getPoint(cell[j],x);						//     Get vertex coordinates of j-th vertex of i-th cell		
			float* point_normal = point_normals.getC0Vector(cell[j]);

			glNormal3f(point_normal[0],point_normal[1],point_normal[2]);
			glVertex3f(x[0],x[1],x[2]);					//     Pass this coordinate to OpenGL
		}	
		glEnd();                                        //4.   Finish rendering current cell
	}
}




void MeshRenderer::drawC1Cells(UnstructuredGrid3D& g)	//Draw 'g' with smooth shading (bilinear interpolation of shading at vertices)
{	
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);			//1. Render cells filled
	glEnable(GL_COLOR_MATERIAL);
  	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glShadeModel(GL_SMOOTH);
	
	
	ScalarAttributes& scalars = g.pointScalars();		//     Get scalar range, to normalize scalar-values next in [0,1]
	float scalar_min = scalars.getMin();
	float scalar_max = scalars.getMax();	
	float scalar_range = scalar_max-scalar_min;
	if (scalar_range==0) scalar_range=1;
	VectorAttributes& point_normals = g.getPointNormals();

	for(int i=0;i<g.numCells();++i)						//2. Draw all cells in the grid 
	{
		int   cell[10];									//
		
		int nvertices = g.getCell(i,cell);				//3.   Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:
		

		for(int j=0;j<nvertices;++j)					//4.   Render current cell
		{
			float x[3];									//     Get vertex coordinates of j-th vertex of i-th cell
			g.getPoint(cell[j],x);						//     Next, get vertex normal of the same vertex			
			float* point_normal = point_normals.getC0Vector(cell[j]);	
			float  scalar = (scalars.getC0Scalar(cell[j])-scalar_min)/scalar_range;
			float R,G,B;								//5.   Get vertex scalar, map it to color
			rainbow(scalar,R,G,B);

			glColor3f(R,G,B);		
			glNormal3f(point_normal[0],point_normal[1],point_normal[2]);
			glVertex3f(x[0],x[1],x[2]);					//6.   Pass the vertex normal, color, and coordinates to OpenGL
		}	
		glEnd();                                        //7.   Finish rendering current cell
	}
}



void MeshRenderer::drawC1TexCells(UnstructuredGrid3D& g)//Draw 'g' by showing C1-interpolated scalars over cells, using textures.
{														//Compare this with drawC1Cells(), which interpolates colors rather than scalars.
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);			//1.   Set basic OpenGL rendering properties.
	glEnable(GL_COLOR_MATERIAL);						//     We use lighting, and set the object's color to pure white. This is because
  	glEnable(GL_LIGHTING);								//     we next use texture-mapping with the GL_MODULATE mode. This effectively
	glEnable(GL_LIGHT0);								//     multiplies a shading value (ranging between 0 and 1) with the hue
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);	//	   delivered by the color map.	
	glShadeModel(GL_SMOOTH);
	glColor3f(1,1,1);	
			
	glEnable(GL_TEXTURE_1D);							//2.   Activate OpenGL's 1D texture mapping. Use the 1D-texture corresponding 
														//     to our colormap.
	glBindTexture(GL_TEXTURE_1D,textureID);	
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

	ScalarAttributes& scalars = g.pointScalars();		//3.    Get scalar range, to normalize scalar-values next in [0,1]
	float scalar_min = scalars.getMin();
	float scalar_max = scalars.getMax();
	VectorAttributes& point_normals = g.getPointNormals();
	
	for(int i=0;i<g.numCells();++i)						//4. Draw all cells in the grid filled bilinearly-interpolated polygons
	{
		int   cell[10];									//Should have enough storage space for our largest cell
		
		int nvertices = g.getCell(i,cell);				//4.1. Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:
		
		for(int j=0;j<nvertices;++j)					//4.2. Render current cell
		{
			float x[3];
			g.getPoint(cell[j],x);						//4.3. Get vertex coordinates of j-th vertex of i-th cell	
			float* point_normal = point_normals.getC0Vector(cell[j]);							
			float value = (g.pointScalars().getC0Scalar(cell[j])-scalar_min)/(scalar_max-scalar_min);
														//4.4. Get vertex scalar value, normalize it in [0,1]    	
			glTexCoord1f(value);						//4.5. Use normalized scalar as texture-coordinate
			glNormal3f(point_normal[0],point_normal[1],point_normal[2]);
			glVertex3f(x[0],x[1],x[2]);					//4.6. Pass vertex coordinate to OpenGL
		}	
		glEnd();                                        //4.7. Finish rendering current cell
	}
	
	glDisable(GL_TEXTURE_1D);							//5.   Done with using 1D textures
}



void MeshRenderer::drawVectors(UnstructuredGrid3D& g)	//Draw vector field on 'g', using simple line glyphs
{														//Since we assume these are eigenvectors, we draw symmetric
	VectorAttributes& vectors = g.pointVectors();		//line glyphs, centered at the grid vertices
	ScalarAttributes& scalars = g.pointScalars();
	
	float S = (normalize)? scale: scale/scalars.getMax();//Compute scale factor for vector glyphs, so that the
														 //longest glyph is 'scale' world-space units. Here we
														 //use the fact that the grid scalars encode the vectors'
														 //lengths.

	glEnable(GL_BLEND);									 //Use blending to make glyphs transparent at both ends
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);	 //and black-and-opaque in the center
	glColor3f(0,0,0);
	glBegin(GL_LINES);
	glDisable(GL_LIGHTING);
	
	for(int i=0;i<g.numPoints();++i)
	{
		Point3d x;
		g.getPoint(i,x.data);							//Get vertex coordinates of i-th vertex	
		Point3d v = vectors.getC0Vector(i);				//Get vector data of i-th vertex			
		
		if (normalize)									//If desired, show normalized vectors
		   v.normalize();
		
		v *= S/2;										//Scale vector to half of the scale-factor S
		
		Point3d v1 = x-v;								//Create a glyph which is of length S, centered at x,
		Point3d v2 = x+v;								//and having the direction of v. We do this since 
														//eigenvectors do not have a meaningful orientation,
														//but just a direction.
		glColor4f(0,0,0,0.1);			
		glVertex3f(v1[0],v1[1],v1[2]);					//Pass the two endpoints of the vector glyph
														//Duplicate the glyph midpoint, since we draw each glyph	
		glColor4f(0,0,0,1);								//as two line segments. Also, make the glyph's endpoints
		glVertex3f(x[0],x[1],x[2]);						//transparent and the glyph center fully opaque

		glColor4f(0,0,0,1);			
		glVertex3f(x[0],x[1],x[2]);

		glColor4f(0,0,0,0.1);			
		glVertex3f(v2[0],v2[1],v2[2]);					
	}

	glEnd();
	
	glDisable(GL_BLEND);
}

