#include "UnstructuredGrid3D.h"
#include <iostream>
#include <map>
#include <deque>
#include <tr1/unordered_set>


void UnstructuredGrid3D::getPoint(int i,float* p)
{
	p[0] = pointsX[i];
	p[1] = pointsY[i];
	p[2] = pointsZ[i];
}


void UnstructuredGrid3D::setPoint(int i,float* p)
{
	pointsX[i] = p[0];
	pointsY[i] = p[1];
	pointsZ[i] = p[2];
}

void UnstructuredGrid3D::normalize()						//Normalize the grid in the [-1,1] cube
{
	float minX= 1.0e6,minY= 1.0e6,minZ= 1.0e6;
	float maxX=-1.0e6,maxY=-1.0e6,maxZ=-1.0e6;
	
	for(int i=0;i<numPoints();++i)							//1. Determine the bounding-box of all points in the grid
	{
		float p[3];
		getPoint(i,p);
		minX = min(p[0],minX); maxX = max(p[0],maxX);
		minY = min(p[1],minY); maxY = max(p[1],maxY);
		minZ = min(p[2],minZ); maxZ = max(p[2],maxZ);
	}
	
	float sizeX = maxX-minX;								//2. Compute a single scaling factor that best fits the grid
	sizeX = (sizeX)? 1/sizeX : 1;							//   in the [-1,1] cube. Using a single factor for x,y, and z
	float sizeY = maxY-minY;								//   ensures that the object is scaled while keeping its
	sizeY = (sizeY)? 1/sizeY : 1;							//   aspect ratio.
	float sizeZ = maxZ-minZ;
	sizeZ = (sizeZ)? 1/sizeZ : 1;
	
	float scale = min(sizeX,min(sizeY,sizeZ));
	
	for(int i=0;i<numPoints();++i)							//3. Use the scaling factor computed above to scale all grid
	{														//   points in the [-1,1] cube
		float p[3];
		getPoint(i,p);
		
		p[0] = 2*((p[0]-minX)*scale-0.5);
		p[1] = 2*((p[1]-minY)*scale-0.5);
		p[2] = 2*((p[2]-minZ)*scale-0.5);
		
		setPoint(i,p);
	}
}


Point3d cellNormal(int size,Point3d* p)						//Compute the normal of a cell whose three vertices are given in p[]
{
	Point3d edge1 = p[1]-p[0];								//We assume that the cell is a triangle. Then, the normal is the
	Point3d edge2 = p[2]-p[1];								//normalized (unit length) value of the cross-product of two cell edges.
	Point3d normal = edge1.cross(edge2);
	normal.normalize();
	return normal;
}	
	
	
void UnstructuredGrid3D::computeFaceNormals()				//Compute face normals for the grid. For this, we use the cellNormal()
{															//function presented above, for all grid triangles.
	for(int i=0;i<numCells();++i)
	{
		int cell[10];
		int size = getCell(i,cell);
		
		Point3d points[10];
		for(int j=0;j<size;++j)
		{
			float p[3];
		    getPoint(cell[j],p);
			points[j] = Point3d(p);
		}
		
		Point3d face_normal = cellNormal(size,points);
		
		faceNormals.setC0Vector(i,face_normal.data);
	}	
}	


void UnstructuredGrid3D::computeVertexNormals()				//Compute vertex normals for the grid. For this, we add, to each vertex,
{															//the normals of all cells that use that vertex. Next, we normalize the result.
	for(int i=0;i<numCells();++i)
	{
		int cell[10];
		int size = getCell(i,cell);

		Point3d face_normal = faceNormals.getC0Vector(i);
		
		for(int j=0;j<size;++j)
		{
			Point3d point_normal = pointNormals.getC0Vector(cell[j]);
			
			point_normal += face_normal;
			
			pointNormals.setC0Vector(cell[j],point_normal.data);
		}
	}
	
	for(int i=0;i<numPoints();++i)
	{
		Point3d point_normal = pointNormals.getC0Vector(i);
		point_normal.normalize();
		pointNormals.setC0Vector(i,point_normal.data);
	}
}

void UnstructuredGrid3D::computePointTriangles()			//Compute map from vertices to all triangles using each vertex
{
	point2cells.resize(numPoints());						//The map has as many entries as vertices in the grid
	
	for(int i=0;i<numCells();++i)							//For each cell, add the cell to the point2cells[] entry of all its vertices
	{
		int cell[10];
		int C = getCell(i,cell);
		for(int c=0;c<C;++c)
		   point2cells[cell[c]].insert(i);
	}
}

void UnstructuredGrid3D::closest(float radius,int pid,int maxNeighbors,vector<int>& result)		
{																				//Given a vertex-index 'pid' in the current mesh, returns all vertices
																				//closest to 'pid' than a given distance 'radius', which are connected
																				//to 'pid' by a path on the mesh-surface located in the 3D ball 
																				//of radius 'radius'

	multimap<float,int>		Q;													//Queue storing the vertices being traversed from 'pid' outwards
	tr1::unordered_set<int> visited;											//Stores the set of already-visited vertices

	Q.insert(make_pair(0,pid));													//Start visiting the mesh from vertex 'pid'
	
	Point3d p;
	int cell[10];
	getPoint(pid,p);
	float radius2 = radius*radius;												//Compute squared radius; faster to test distances against this 
	
	result.clear();																//Make sure that the result[] container is empty and
	result.reserve(maxNeighbors);												//will not need extra expensive reallocations
	
	while(!Q.empty() && result.size()<maxNeighbors)								//Search in distance-order from 'pid', until we reach either the
	{																			//'radius' distance limit or the maximum allowed number of neighbors
		int pid = Q.begin()->second; Q.erase(Q.begin());
		result.push_back(pid);													//Get the top of the queue: this is a valid neighbor, so save it

		const UnstructuredGrid3D::CellsOfPoint& cells = cellsOfPoint(pid);		//Get all grid-cells using 'pid'

		for(UnstructuredGrid3D::CellsOfPoint::const_iterator it=cells.begin(),ie=cells.end();it!=ie;++it)
		{																		//See which points of the above cells qualify for neighbors of 'pid'
			int C = getCell(*it,cell);											//For this, test all vertices of these cells against our criteria,
			for(int i=0;i<C;++i)												//and if they pass, add them to the queue
			{	
				int qid = cell[i];
				Point3d q;
				getPoint(qid,q.data);	
				float dist2 = q.dist2(p);	
				if (dist2>radius2) continue;									//Cell-vertex too far? If so, skip it
				if (visited.find(qid)!=visited.end()) continue;					//Cell-vertex already encountered? If so, skip it
				
				visited.insert(qid);											//Mark cell-vertex as being visited (encountered)
				Q.insert(make_pair(dist2,qid));									//Add cell-vertex for traversal	
			}
		}		
	}
}



