#include <GLUT/glut.h>									//GLUT library
#include <iostream>
#include "MeshPCA.h"
#include "MeshRenderer.h"
#include "UnstructuredGrid3D.h"
#include "PlyReader.h"
#include "zpr.h"


UnstructuredGrid3D*			grid			= 0;		//The mesh to visualize
MeshRenderer*				mesh_renderer   = 0;		//A special renderer to draw a mesh with classifier values
MeshPCA*					pca				= 0;		//Engine to compute curvature tensor on a 3D mesh
int							show_metric     = 0;		//Which classifier values to show (0=Pauly,1=Westin,2=Clarenz)
ScalarAttributes*			classifiers[3];				//All classifier values being computes from the curvature tensor
float						radius			= 0.05;		//Size of neighborhood (world units) on which to estimate eigen-components
bool						search_on_surface = false;	//Search neighborhoods strictly on the mesh surface (slower, more accurate)
														//or in 3D space (faster, less accurate). See MeshPCA class.
MeshRenderer::DRAW_STYLE	draw_style		= MeshRenderer::DRAW_C1_TEX_CELLS;
														//How to draw the mesh



static void mouseclick(int button, int state, int x, int y)
{
	int keys = glutGetModifiers();						//The mouse events are interpreted as follows:	
	if (keys & GLUT_ACTIVE_CTRL)						// 
	button = GLUT_MIDDLE_BUTTON;						// -left button + move:                         rotate viewpoint
	if (keys & GLUT_ACTIVE_SHIFT)						// -middle button (or left button+Control key): zoom in / out
	button = GLUT_RIGHT_BUTTON;							// -right buttom (or left button+Shift key):    translate viewpoint

	zprMouse(button,state,x,y);							//Use the ZPR library to manipulate the viewpoint. The library sets the modelview  
														//OpenGL matrix based on the mouse events, thereby changing the current viewpoint.
}


static void mousemotion(int x, int y)
{
	zprMotion(x,y);
	glutPostRedisplay();
}


static void viewing(int width, int height)					//GLUT window-reshape callback
{
	glutReshapeWindow(width,height);  
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70,double(width)/height,0.5,7);
	glViewport(0,0,width,height);
}


void showComponent(int classifier)							//Select which classifier values to display next
{															//Also print some statistics on the shown classifier
	grid->pointScalars() = *classifiers[classifier];

	cout<<"Showing anisotropy metric: "
		<<((show_metric==0)?"Pauly's variation":(classifier==1)?"Westin's planarity":"Clarenz's momentum")
		<<" (radius = "<<radius<<")"
		<<". Classifier range: "<<classifiers[classifier]->getMin()<<", "<<classifiers[classifier]->getMax()<<endl;	
	cout<<"Neighborhood computation: "<<((search_on_surface)?"on surface":"in 3D space")<<endl;	
}


static void keyboard(unsigned char ch,int,int)				//GLUT keyboard callback
{
	switch(ch)
	{
	case ' ':												//Cycle through the classifiers to show
		show_metric = (++show_metric) % 3;
		showComponent(show_metric);
		break;
	case '+':												//Increase radius of neighborhood used for PCA computation
	case '=':
		radius *= 1.2;
		pca->compute(radius);
		showComponent(show_metric);
		break;	
	case '-':												//Decrease radius of neighborhood used for PCA computation
	case '_':
		radius /= 1.2;
		pca->compute(radius);
		showComponent(show_metric);
		break;	
	case 'd':												//Cycle through the drawing styles for the mesh
	case 'D':
		draw_style = (MeshRenderer::DRAW_STYLE)((++int(draw_style)) % 3);
		mesh_renderer->setDrawingStyle(draw_style);	
		break;
	case 'r':												//Reset the viewpoint
	case 'R':
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		zprInit(0,0,-2);
		break;	
	case 's':
	case 'S':
		search_on_surface = !search_on_surface;
		pca->searchOnSurface(search_on_surface);
		pca->compute(radius);
		showComponent(show_metric);
		break;			
	}
	
	glutPostRedisplay();	
}		


static void display()										//GLUT display callback
{
   glMatrixMode(GL_MODELVIEW);	
   glPushMatrix();
   glTranslatef(0,0,-2);									//Translate the mesh so it's in front of the near clipping plane
   mesh_renderer->draw(*grid);								//Draw the mesh
   glPopMatrix();
}



int main(int argc,char* argv[])								//Main program
{
	cout<<"2. Surface curvature-tensor anisotropy metrics"<<endl;
	cout<<"In the graphics window, press"<<endl;
	cout<<"    space: cycle through the 3 anisotropy metrics to show:"<<endl;
	cout<<"           - Pauly's surface variation metric;"<<endl;
	cout<<"           - Westin's planar anisotropy metric;"<<endl;
	cout<<"           - Clarenz's moment classifier metric."<<endl;
	cout<<"    +,-:   increase/decrease neighborhood-radius for PCA computation"<<endl;
	cout<<"    s,S:   search neighbors strictly on mesh surface (slower, more accurate)"<<endl;
	cout<<"           or in 3D space (faster, less accurate)"<<endl;
	cout<<"    d,D:   cycle through the mesh drawing styles"<<endl;
	cout<<"    r,R:   reset viewpoint"<<endl;

	const char* filename = (argc>1)? argv[1]:"DATA/bunny.ply";
	
	glutInit(&argc, argv);									//1. Initialize the GLUT toolkit
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE|GLUT_DEPTH);	//2. Ask GLUT to create next windows with a RGB framebuffer and a Z-buffer too
	glutInitWindowSize(500,500);							//3. Tell GLUT how large are the windows we want to create next
	glutCreateWindow("1. Surface curvature tensors");		
	zprInit(0,0,-2);

	PlyReader rdr;											//4.  Read a 3D mesh stored in a file in the PLY format
    grid = rdr.read(filename);
    grid->normalize();										//5.  Normalize the mesh in the [-1,1] cube. This makes various parameter settings easier next.
	grid->computeFaceNormals();								//    Compute mesh face and vertex normals, used for shading
	grid->computeVertexNormals();
	grid->computePointTriangles();
	
	pca = new MeshPCA(*grid,200);							//6.  Create an engine for PCA computation of the curvature tensor.
	pca->searchOnSurface(search_on_surface);				//    Tell if we search point-neighbors on mesh surface or in 3D space.
	pca->compute(radius);									//    Initialize the neighborhood-radius and maximal number of neighbors
															//    for this engine.
	
	classifiers[0]  = &pca->variation; classifiers[1]  = &pca->planarity; classifiers[2]  = &pca->classifier;
															//7.  Get hold of the classifier values that the engine will compute next.
																
	mesh_renderer = new MeshRenderer();						//8.  Create a renderer for the mesh and associated classifiers
	mesh_renderer->setDrawingStyle(draw_style);				//    Set its various parameters to our desired values
	mesh_renderer->showVectors(false);						//    Do not show eigenvectors, we only show classifier-metrics in this sample
	showComponent(show_metric);								//9.  We visualize the first classifier metric next
					
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouseclick);
	glutMotionFunc(mousemotion);
	glutReshapeFunc(viewing);
	glutDisplayFunc(display);
	
	glutMainLoop();

	return 0;
}




